﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Collections.Specialized;
using System.IO;
using OSGeo.MapServer;
using System.Configuration;
using bars.global.functions;
using bars.global.enums;
using bars.system.extensions;

using System.Diagnostics;

namespace bars.web.mapserver
{
    /// <summary>
    /// Map generation class
    /// </summary>
    public class MapFactory
    {
        private const string DEFAULT_WMS_VERSION = "1.1.1";
        private const string SQLSERVER_PLUGIN_DLL = "msplugin_mssql2008.dll";
        private const string MAPFILE_MACRO_PLUGINS_SQLSERVER = "sqlplugin";

        private string _pluginLibraryPath = string.Empty;
        private string _mapFile = string.Empty;
        private string _shapeFilePath = string.Empty;
        private string _SQLConnectionString = string.Empty;
        private string _projectionLibraryPath = string.Empty;
        private string _templatePath = string.Empty;


        private string _WMSonlineResource = string.Empty;

        private EventLog _logInstance = null;


		  
        private EventLog Log
        {
            get
            {
                if (_logInstance == null)
                {
                    _logInstance = new EventLog();
                    _logInstance.Source = "BARS_MapFactory";
                }

                return _logInstance;
            }
        }

        /// <summary>
        /// Gets or sets the template path.
        /// </summary>
        /// <value>The template path.</value>
        public string TemplatePath
        {
            get 
            { 
                return _templatePath; 
            }
            set 
            { 
                _templatePath = value; 
            }
        }

        /// <summary>
        /// Gets or sets the WMS online resource URL.
        /// </summary>
        /// <value>The WMS online resource URL.</value>
        public string WMSOnlineResourceURL
        {
            get
            {
                return _WMSonlineResource;
            }
            set
            {
                _WMSonlineResource = value;
            }
        }

        /// <summary>
        /// Gets or sets the projection library path.
        /// </summary>
        /// <value>The projection library path.</value>
        public string ProjectionLibraryPath
        {
            get
            {
                return _projectionLibraryPath;
            }

            set
            {
                if (value != null)
                {
                    _projectionLibraryPath = value.EndsWith("\\") ? value : value + "\\";
                }
            }
        }

        /// <summary>
        /// Gets or sets the map file path.
        /// </summary>
        /// <value>The map file path.</value>
        private string MapFile
        {
            get { return _mapFile; }
            set { _mapFile = value; }
        }

        /// <summary>
        /// Gets or sets the SQL connection string that should be used by the mapserver api.
        /// </summary>
        /// <value>The SQL connection string.</value>
        public string SQLConnectionString
        {
            get
            {
                return _SQLConnectionString;
            }
            set
            {
                _SQLConnectionString = value;
            }
        }

        /// <summary>
        /// Gets or sets the plugin file path.
        /// </summary>
        /// <value>The plugin file path.</value>
        public string PluginLibraryPath
        {
            get
            {
                return _pluginLibraryPath;
            }

            set
            {
                if (value != null)
                {
                    _pluginLibraryPath = value.EndsWith("\\") ? value : value + "\\";
                }
            }
        }

        /// <summary>
        /// Gets or sets the shape file path.
        /// </summary>
        /// <value>The shape file path.</value>
        public string ShapeFilePath
        {
            get
            {
                return _shapeFilePath;
            }
            set
            {
                if (value != null)
                {
                    _shapeFilePath = value.EndsWith("\\") ? value : value + "\\";
                }
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="MapFactory"/> class.
        /// </summary>
        /// <param name="mapFilePath">The map file path.</param>
        public MapFactory(string mapFilePath)
        {
            if (mapFilePath == string.Empty)
            {
                throw new ArgumentException("You must provide a mapfile when instantiating an instance of bars.web.mapserver.MapManager", "mapFilePath");
            }

            MapFile = mapFilePath;
        }

        /// <summary>
        /// Gets the feature info from WMS.
        /// </summary>
        /// <param name="WMSqueryString">The WM squery string.</param>
        /// <returns>A string containing the processed templates for the queried layers.</returns>
        /// <exception cref="ArgumentNullException">Thrown if the X or Y point values are not supplied in the WMS GetFeatureInfo request</exception>
        /// <exception cref="ArgumentException">Thrown in the X or Y point values supplied in the WMS GetFeatureInfo request cannot be converted to valid doubles</exception>
        public string GetFeatureInfoFromWMS(NameValueCollection WMSqueryString)
        {
            //Set the projection library environment variable used by mapscript.dll
            Environment.SetEnvironmentVariable("PROJ_LIB", ProjectionLibraryPath);

            string output = string.Empty;

            try
            {
                using (mapObj map = new mapObj(MapFile))
                {
                    ProcessLayers(map, WMSqueryString);

                    map.web.metadata.set("wms_onlineresource", WMSOnlineResourceURL);

                    string xVal = WMSqueryString["X"];
                    string yVal = WMSqueryString["Y"];

                    if (xVal == null || yVal == null)
                    {
                        throw new ArgumentNullException("The X or Y point value has not been suppplied in the GetFeatureInfo request");
                    }

                    double pointX = 0.0;
                    double pointY = 0.0;

                    try 
	                {	        
                        pointX = Convert.ToDouble(xVal);
                        pointY = Convert.ToDouble(yVal);
	                }
	                catch (Exception e)
	                {
		                throw new ArgumentException("The X or Y point value supplied in the GetFeatureInfo request is not a valid decimal",e);
	                }

                    string layersQS = WMSqueryString["QUERY_LAYERS"];

                    if (layersQS == null)
                    {
                        throw new ArgumentNullException("The QUERY_LAYERS parameter of the WMS GetFeatureInfo request is not specified correctly");
                    }

                    //Load the parameters from the wms request into the map
                    using (OWSRequest request = new OWSRequest())
                    {
                        for (int i = 0; i < WMSqueryString.Count; i++)
                        {
                            request.setParameter(WMSqueryString.GetKey(i), WMSqueryString.Get(i));
                        }

                        string wmsVersion = WMSqueryString["VERSION"];

                        if (wmsVersion == null || wmsVersion == string.Empty) wmsVersion = DEFAULT_WMS_VERSION;

                        map.loadOWSParameters(request, wmsVersion);
                    }

                    //Reporject X & Y pixel co-ordinates in map co-ordintes.
                    double minX = map.extent.minx;
                    double maxX = map.extent.maxx;
                    double geoX = minX + ((pointX / (double)map.width) * (maxX - minX));

                    double minY = map.extent.miny;
                    double maxY = map.extent.maxy;
                    double geoY = maxY - ((pointY / (double)map.height) * (maxY - minY));

                    string[] queryLayers = layersQS.Split(',');

                    using (pointObj point = new pointObj(geoX, geoY, 0, 0))
                    {
                        foreach (string layerName in queryLayers)
                        {
                            using (layerObj layer = map.getLayerByName(layerName))
                            {
                                int queryResult = layer.queryByPoint(map, point, (int)MS_QUERY_MODE.MS_QUERY_SINGLE, -1);
                            }
                        }
                    }

                    map.prepareQuery();

                    string[] names = { "Token1" };
                    string[] values = { "Value1" };

                    output = map.processQueryTemplate(names, values, 10);
                }
                return output;
            }
            catch (Exception ex)
            {
                throw;
            }
        }

        /// <summary>
        /// Gets the map image from WMS get request.
        /// </summary>
        /// <param name="WMSqueryString">The query string from the WMS request.</param>
        /// <returns>A stream containing the map image</returns>
        public Stream GetMapFromWMS(NameValueCollection WMSqueryString)
        {
            string wmsLogMessage = "Recieved WMSQuery " + WMSqueryString.ToString() + '\n'
                + "MapfilePath: " + MapFile + '\n'
                + "PluginLibPath: " + PluginLibraryPath + '\n'
                + "ProjectionLibraryPath: " + ProjectionLibraryPath + '\n'
                + "WMSOnlineResourceURL" + WMSOnlineResourceURL + '\n'
                + "ShapeFilePath" + ShapeFilePath;

            Log.WriteEntry(wmsLogMessage, EventLogEntryType.Information);
            

            //Set the projection library environment variable used by mapscript.dll
            Environment.SetEnvironmentVariable("PROJ_LIB", ProjectionLibraryPath);

            try
            {
                using (mapObj map = new mapObj(MapFile))
                {
                    ProcessLayers(map, WMSqueryString);
                    Log.WriteEntry("Initialised map instance from mapfile", EventLogEntryType.SuccessAudit);
                    
                    map.web.metadata.set("wms_onlineresource", WMSOnlineResourceURL);

                    MemoryStream outputStream = null;

                    using (OWSRequest request = new OWSRequest())
                    {
                        for (int i = 0; i < WMSqueryString.Count; i++)
                        {
                            request.setParameter(WMSqueryString.GetKey(i), WMSqueryString.Get(i));
                        }
                        Log.WriteEntry("Initialised WMS query from querystring", EventLogEntryType.SuccessAudit);

                        string wmsVersion = WMSqueryString["VERSION"];

                        if (wmsVersion == null || wmsVersion == string.Empty) wmsVersion = DEFAULT_WMS_VERSION;

                        map.loadOWSParameters(request, wmsVersion);
                        Log.WriteEntry("Configured map from WMS query", EventLogEntryType.SuccessAudit);

                        using (imageObj mapImage = map.draw())
                        {
                            byte[] image = mapImage.getBytes();
                            outputStream = new MemoryStream(image);
                        }
                        Log.WriteEntry("Rendered map image to output stream", EventLogEntryType.SuccessAudit);
                    }
                    return outputStream;
                }
            }
            catch (Exception e)
            {
                //TODO: Redirect messages to a silent logging mech.
                GlobalFunctions.RaiseNotification(e.Message, GlobalEnums.NotificationType.Error);
                wmsLogMessage = "Exception raised when generating map for wms query"
                    + WMSqueryString.ToString() + '\n'
                    + "Exception: " + e.Message;

                if (e.InnerException != null)
                {
                    wmsLogMessage = wmsLogMessage + '\n'
                        + "Inner Exception: " + e.InnerException.Message;
                }

                Log.WriteEntry(wmsLogMessage, EventLogEntryType.Error);

                e.SetNotificationRaised(true);

                throw;
            }

        }

        /// <summary>
        /// Configures dynamic properties on each map layer defined in the mapfile.
        /// </summary>
        /// <param name="map">The map.</param>
        private void ProcessLayers(mapObj map, NameValueCollection WMSqueryString)
        {
            //All layers must be processsed including those that aren't requested otherwise exceptions are thrown - notes Case 2991 -  8/02/2011 - 11:14
            for (int layerIndex = 0; layerIndex < map.numlayers; layerIndex++)
            {
                using (layerObj layer = map.getLayer(layerIndex))
                {
                    ProcessLayer(WMSqueryString, layer);
                }
            }
        }

        private void ProcessLayer(NameValueCollection WMSqueryString, layerObj layer)
        {
            if (layer.plugin_library != null && layer.plugin_library.ToLower() == MAPFILE_MACRO_PLUGINS_SQLSERVER)
            {
                if (PluginLibraryPath == null || PluginLibraryPath == string.Empty)
                {
                    throw new InvalidOperationException(String.Format("The map file {0} or one of its includes contains layers that use the MapServer SQL plugin dll ({1}) but the MapManager.PluginLibraryPath has not been set.", MapFile, MAPFILE_MACRO_PLUGINS_SQLSERVER));
                }

                if (SQLConnectionString == null || SQLConnectionString == string.Empty)
                {
                    throw new InvalidOperationException(String.Format("The map file {0} or one of its includes contains layers that use the MapServer SQL plugin dll but the MapManager.SQLConnectionString property has not been set.", MapFile));
                }

                string filterString = layer.getFilterString();

                if (filterString != null && filterString != string.Empty)
                {
                    filterString = filterString.Replace("\"", string.Empty);
                    MapFilterFactory filterFactory = new MapFilterFactory(filterString);
                    layer.setFilter(filterFactory.GetFilterTerm(WMSqueryString));
                }

                layer.plugin_library = PluginLibraryPath + SQLSERVER_PLUGIN_DLL;

                if (layer.connection == null || layer.connection == string.Empty)
                {
                    layer.connection = SQLConnectionString;
                    layer.connectiontype = MS_CONNECTION_TYPE.MS_PLUGIN;
                }

            }

            //Set the path for the shapefile.
            if (layer.connectiontype == MS_CONNECTION_TYPE.MS_SHAPEFILE || layer.connectiontype == MS_CONNECTION_TYPE.MS_TILED_SHAPEFILE)
            {
                if (ShapeFilePath == null || ShapeFilePath == string.Empty)
                {
                    throw new InvalidOperationException(String.Format("The map file {0} or one of its includes contains layers that use shapefiles but the MapManager.ShapeFilePath has not been set.", MapFile));
                }

                if (layer.data != null || layer.data != string.Empty)
                {
                    layer.data = ShapeFilePath + layer.data.TrimStart('\\');
                }
            }

            //Set the path for the template
            if (layer.template != null && layer.template != string.Empty)
            {
                layer.template = TemplatePath + layer.template.TrimStart('\\');
            }
        }
    }
}
